import argparse
import asyncio

from .youtube import Playlist, Video
from .streamdj import StreamDj
from .exceptions import StreamDjException


class Ui:
    """
    Class with user interface
    """

    def __init__(self):
        self.parser = argparse.ArgumentParser(description="Easy way to use stream dj")

        self.parser.add_argument(
            "user", type=str, help="stream dj user name you want send music to"
        )
        self.parser.add_argument(
            "-v", "--video", type=str, metavar="URL", help="send youtube video"
        )
        self.parser.add_argument(
            "-p",
            "--playlist",
            type=str,
            metavar="URL",
            help="youtube playlist, all vidios from it will be send to stream dj",
        )
        self.parser.add_argument(
            "-P",
            "--playlistsearch",
            type=str,
            metavar="TEXT",
            help="same as --playlist, but it'll search for a playlist in youtube",
        )
        self.parser.add_argument(
            "-d",
            "--delay",
            type=float,
            metavar="SECONDS",
            help="delay between sending videos from playlist (float) defalut=0",
            default=0.0,
        )
        self.parser.add_argument(
            "-q",
            "--quantity",
            action="store_true",
            help="show quantity of tracks in stream dj.",
        )
        self.parser.add_argument(
            "-l", "--list", action="store_true", help="show all tracks in stream dj"
        )
        self.parser.add_argument(
            "-s",
            "--skip",
            action="store_true",
            help="(testing) send requests for skiping now playing track in stream dj",
        )
        self.parser.add_argument(
            "-a",
            "--author",
            type=str,
            metavar="NAME",
            help="author name of track sender",
        )
        self.args = self.parser.parse_known_args()[0]
        self.dj = StreamDj(self.args.user, self.args.author)

        self._threads = []
        self._is_sending_ended = False

    async def run(self):
        await self.dj.update_channel_id()
        if self.args.quantity or self.args.list or self.args.skip:
            tracks = await self.dj.track_list()
            if self.args.quantity:
                quantity = len(tracks)
                print(f"\n\ntracks quantity: {quantity}\n\n")
            if self.args.list:
                for track in tracks:
                    print(f"{track.author}: {track.title}\n{track.id=} {track.skip=}\n")
            if self.args.skip and tracks:
                await self.dj.update_proxy_list()
                track_id = tracks[0].id
                tasks = []
                for _ in range(await self.dj.get_proxy_amount()):
                    tasks.append(asyncio.create_task(self._vote_skip_and_print_result(track_id)))
                for task in tasks:
                    await task
        if self.args.video:
            await self._send_request_and_print_result(Video("Video", self.args.video))
        if self.args.playlist or self.args.playlistsearch:
            if self.args.playlistsearch:
                self.args.playlist = await self._chouse_playlist()
            tracks = await Playlist(self.args.playlist).get_videos()
            print(f"\n\nVideos fetched: {len(tracks)}\n\n")
            tasks = []
            for track in tracks:
                tasks.append(asyncio.create_task(self._send_request_and_print_result(track)))
                await asyncio.sleep(self.args.delay)

            for task in tasks:
                await task

    async def _chouse_playlist(self) -> Playlist:
        text = self.args.playlistsearch
        lists = await Playlist.search(text)
        print()
        for i, lst in enumerate(lists):
            print(f"[{i}] - videos({lst.videos_amount}): {lst.name}")
        try:
            i = int(input("chouse (blank or trash=0): "))
        except:
            i = 0
        return lists[i].url

    async def _send_request_and_print_result(self, video: Video):
        try:
            await self.dj.send(video.url)
            print(f"success sended: {video.title}")
        except StreamDjException as e:
            print(f"error while sending: {video.title=} {e=}")

    async def _vote_skip_and_print_result(self, track_id: int):
        try:
            await self.dj.vote_skip(track_id)
            print(f"success skip request: {track_id}")
        except StreamDjException as e:
            print(f"error while skiping: {track_id=} {e=}")



import re
from collections import namedtuple

import aiohttp
from bs4 import BeautifulSoup


PlaylistSearchResult = namedtuple("PlaylistSearchResult", ("name", "url", "videos_amount"))
Video = namedtuple("Video", ("title", "url"))


class Playlist:
    """
    class to work with youtube playlist

    Args:
        list_id_or_url: str - youtube playlist id or url with playlist id
        for example https://www.youtube.com/playlist?list=AAA_AAAAAA-AAAAAAA is url
        AAA_AAAAAA-AAAAAAA is playlist id

    Examples:
        >>> y = Playlist('LISTID')
        >>> y.get_videos()
        [Video(title='Some video 1', url='https://www.youtube.com/watch?v=1111'),
         Video(title='Some video 2', url='https://www.youtube.com/watch?v=2222')]
    """

    _playlist_search_url_template = (
        "https://yewtu.be/search?q={text} content_type%3Aplaylist"
    )
    _playlist_url_template = "https://yewtu.be/playlist?list={list_id}&page={page}"
    _video_url_template = "https://www.youtube.com{href}"
    _video_from_playlist_re = re.compile(
        r"<a style=\"width:100%\" href=\"(/watch\?v=[^&]+).*\"\>[^`]+?<p dir=\"auto\">(.*)</p>"
    )
    _video_from_mix_re = re.compile(r"<a href=(\"\/watch\?v=[^&]+)[^`]+?<p dir=\"auto\">(.+)<\/p>")

    def __init__(self, list_id_or_url: str):
        if "list=" in list_id_or_url:
            regex_result = re.search(r"list=([a-zA-Z0-9-_]+)", list_id_or_url)
            if regex_result is None:
                raise ValueError(f"invalid argument {list_id_or_url=}")
            self._list_id = regex_result.group(1)
        else:
            self._list_id = list_id_or_url

    @staticmethod
    async def search(text: str) -> list[PlaylistSearchResult]:
        """search for a youtube playlists and return list of them"""

        url = Playlist._playlist_search_url_template.format(text=text)
        async with aiohttp.request("GET", url) as response:
            if response.status // 100 != 2:
                raise ConnectionError(f"search {response.status=} {url=}")
            html = await response.text()
            bs = BeautifulSoup(html, "html.parser")
            divs = bs.find_all(
                "div",
                attrs={
                    "class": "pure-u-1 pure-u-md-1-4",
                },
            )

        results = []
        for div in divs:
            a = div.div.a
            url = a.attrs["href"]
            p_amount, p_name = a.find_all("p")
            amount = int(p_amount.text.split()[0])
            name = p_name.text
            results.append(
                PlaylistSearchResult(
                    name = name,
                    url = url,
                    videos_amount = amount,
                )
            )
        return results

    async def get_videos(self) -> list[Video]:
        """get all videos from playlist"""

        videos = set()
        page = 1

        while True:
            url = self._playlist_url_template.format(list_id=self._list_id, page=page)
            async with aiohttp.request("GET", url) as response:
                if response.status != 200:
                    raise ConnectionError(
                        f"request to {page=} {response.status=} {url=}"
                    )
                html = await response.text()

                videos.update(self._fetch_videos_from_html(html))

                if self._is_next_page_exist(html):
                    page += 1
                else:
                    break

        return list(videos)

    def _fetch_videos_from_html(self, html: str) -> set[Video]:
        videos = set()
        for href, title in self._video_from_playlist_re.findall(html):
            url = self._video_url_template.format(href=href)
            videos.add(Video(title, url))
        if not videos:
            for href, title in self._video_from_mix_re.findall(html):
                url = self._video_url_template.format(href=href)
                videos.add(Video(title, url))
        return videos

    def _is_next_page_exist(self, html: str) -> bool:
        bs = BeautifulSoup(html, "html.parser")
        next_page_div = bs.find(
            "div",
            attrs={
                "class": "pure-u-1 pure-u-lg-1-5",
                "style": "text-align:right",
            },
        )
        return getattr(next_page_div, "a", None) is not None



import asyncio
import errno
import re
import random
from collections import namedtuple
from time import sleep

import aiohttp
import names

from .exceptions import StreamDjException


Track = namedtuple("Video", ("id", "title", "author", "skip"))


class StreamDj:
    """
    class to work with streamdj.app

    Args:
        channel_name: str - nickname of streamdj user
        for example https://streamdj.app/c/e6000000000 is url e6000000000 is channel name

        author_name: str - name that will shows as a sender of music
        if leave it `None` it will generates by names lib. generated names looks like `Victor Owens` `Regina Franks`

    Examples:
        >>> s = StreamDj('someone')
        >>> s.send('https://www.youtube.com/watch?v=POb02mjj2zE')
        {'success': 1}
    """

    _channel_url_template = "https://streamdj.app/c/{name}"
    _track_list_url_template = "https://streamdj.app/includes/back.php?func=playlist&channel={channel_id}&c="
    _send_url_template = (
        "https://streamdj.app/includes/back.php?func=add_track&channel={channel_id}"
    )
    _vote_skip_url = "http://streamdj.app/includes/back.php?func=vote_skip"
    _proxy_list_url = (
        "https://raw.githubusercontent.com/TheSpeedX/PROXY-List/master/http.txt"
    )

    def __init__(self, channel_name: str, author_name: str | None = None):
        self._channel_name = channel_name
        self._custom_author_name = author_name

        self._channel_id: str | None = None
        self._proxies: list[str] | None = None


    async def track_list(self) -> list[Track]:
        """get list of tracks from streamdj"""

        if self._channel_id is None:
            raise ValueError("update channel id before")

        url = self._track_list_url_template.format(channel_id=self._channel_id)

        async with aiohttp.request("GET", url) as response:
            if response.status // 100 != 2:
                raise ConnectionError(
                    f"cant get tracks from {self._channel_name}, status_code={response.status}"
                )

            jsn = await response.json(content_type=None)

        tracks = []
        if not jsn:
            return []

        for i in jsn:
            track = jsn[i]
            tracks.append(
                Track(track["id"], track["title"], track["author"], track["skip"])
            )
        return tracks

    async def send(self, track_url: str):
        """
        send track to streamdj and return a result

        Args:
            track_url: str - youtube video url
            looks like `https://www.youtube.com/watch?v=POb02mjj2zE`

        raise `StreamDjException` if track is not sended
        """

        if self._channel_id is None:
            raise ValueError("update channel id before")

        response = await self._request(
            self._send_url_template,
            {"channel_id": self._channel_id},
            {
                "url": track_url,
                "author": self._author_name,
            },
        )
        error = response.get("error", None)
        if error is not None:
            raise StreamDjException(error)

    async def vote_skip(self, track_id: int):
        if self._proxies is None:
            raise ValueError("update proxy list before")
        if self._channel_id is None:
            raise ValueError("update channel id before")

        if not await self.get_proxy_amount():
            raise StreamDjException("no proxies")

        proxy = random.choice(self._proxies)
        self._proxies.remove(proxy)
        proxy = f"http://{proxy}"
        data = {
            "channel": str(self._channel_id),
            "track_id": str(track_id),
        }
        while True:
            try:
                response = await self._request(self._vote_skip_url, {}, data, proxy)
                error = response.get("error", None)
                if error is not None:
                    raise StreamDjException(error)
            except OSError as e:
                if e.errno != errno.ETOOMANYREFS:
                    raise e
                await asyncio.sleep(10)

    @property
    def _author_name(self):
        if self._custom_author_name is not None:
            return self._custom_author_name
        return names.get_full_name()

    async def _request(
        self, url_template: str, url_params: dict, data: dict, proxy: str | None = None
    ) -> dict:
        url = url_template.format(**url_params)

        async with aiohttp.request("POST", url, data=data, proxy=proxy, timeout=aiohttp.ClientTimeout(total=60)) as response:

            if response.status >= 500:
                sleep(5)
                return await self._request(url_template, url_params, data, proxy)

            if response.status // 100 != 2:
                raise ConnectionError(
                    f"error while sending request {url=} to {self._channel_name=}, {response.status=}."
                )

            try:
                return await response.json(content_type=None)
            except Exception as e:  # if response not in json format
                print(e)
                text = await response.text(encoding="utf-8")
                if "Technical problems, come back later." in text:
                    sleep(5)
                    return await self._request(url_template, url_params, data, proxy)
                else:
                    return {
                        "error": f"Does not sended cuz streamdj.app is great. response: {text}"
                    }

    async def get_proxy_amount(self):
        if self._proxies is None:
            raise ValueError("update proxy list before")
        return len(self._proxies)


    async def update_proxy_list(self):
        """get proxy list from github repository"""

        url = self._proxy_list_url
        async with aiohttp.request("GET", url) as response:


            if response.status // 100 != 2:
                raise ConnectionError(f"cant get proxy list {response.status=}")

            text = await response.text()

        self._proxies = text.splitlines()

    async def update_channel_id(self):
        """get channel id from streamdj"""

        url = self._channel_url_template.format(name=self._channel_name)
        async with aiohttp.request("GET", url) as response:

            if response.status == 404:
                raise ValueError(f"channel with {self._channel_name=} does not exist.")
            elif response.status // 100 != 2:
                raise ConnectionError(f"can't update channel id {response.status=}")

            text = await response.text()

        result = re.search(r"onclick=\"add_track\(([0-9]+)\)\"", text)
        if result is None:
            raise ValueError(f"cant find channel id. {self._channel_name=}")

        self._channel_id = result.group(1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easy-streamdj']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp[speedups]>=3.8.1,<4.0.0',
 'beautifulsoup4>=4.11.1,<5.0.0',
 'names>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'easy-streamdj',
    'version': '1.1.1',
    'description': 'script for sending all youtube playlist to streamdj.app',
    'long_description': 'Works only if music requests is free.\n\n# Easy streamdj.app\nCreated for easyer send/skip tracks in streamdj.app.  \nproxy list for skip tracks from [here](https://github.com/TheSpeedX/PROXY-List)\n\n## How to install\n```bash\npip install easy-streamdj\n```\n\n## Usage examples\nget some help:\n```bash\npython -m easy-streamdj\n```\n\nsend one track:\n```bash\npython -m easy-streamdj some_channel --video "https://www.youtube.com/watch?v=dQw4w9WgXcQ"\n\n```\n\nsend playlist of tracks:\n```bash\npython -m easy-streamdj some_channel --playlist "https://www.youtube.com/playlist?list=PL2aMEXnwcG3nqpr49qfCJ5vLTuxImPdme"\n```\n\nsend playlist first playlist finded on youtube:\n```bash\npython -m easy-streamdj some_channel --playlistserch "good music"\n```\n\nskip current track:\n```bash\npython -m easy-streamdj some_channel --skip\n```\n\nrun easy streamdj over tor (linux):\n```bash\nsudo systemctl start tor  # start tor service if it is not started already\ntorify python -m easy-streamdj some_channel -P "bad music" --delay 12 --author "anonymous"\n```\n\n## Install from source\nInstall [python3.10+](https://www.python.org/downloads/)\n```bash\ngit clone https://github.com/e6000000000/easy-streamdj.git\ncd easy-streamdj\npip install poetry\npoetry build\npip install dist/*.whl\n```\n',
    'author': 'a',
    'author_email': 'skipper224483@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/e9000000000/easy-streamdj',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
